@@ -1,3 +1,4 @@
+{-# OPTIONS -cpp #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Distribution.PackageDescription
@@ -72,9 +73,10 @@ module Distribution.PackageDescription (
         updatePackageDescription,
         -- * Utilities
         ParseResult(..),
-        PError, showError,
+        PError, PWarning, showError,
         hcOptions,
         autogenModuleName,
+        haddockName,
 #ifdef DEBUG
         hunitTests,
         test
@@ -83,8 +85,8 @@ module Distribution.PackageDescription (
 
 import Control.Monad(liftM, foldM, when)
 import Data.Char
-import Data.Maybe(fromMaybe, fromJust, isNothing, catMaybes)
-import Data.List (nub)
+import Data.Maybe(fromMaybe, isNothing, catMaybes)
+import Data.List (nub,lookup)
 import Text.PrettyPrint.HughesPJ
 import System.Directory(doesFileExist)
 import System.Environment(getProgName)
@@ -103,6 +105,7 @@ import Distribution.Simple.Utils(currentDir, die, dieWithLocation, warn)
 import Language.Haskell.Extension(Extension(..))
 
 import Distribution.Compat.ReadP as ReadP hiding (get)
+import Distribution.Compat.FilePath(joinFileExt)
 
 #ifdef DEBUG
 import HUnit (Test(..), assertBool, Assertion, runTestTT, Counts, assertEqual)
@@ -111,7 +114,7 @@ import Distribution.ParseUtils  (runP)
 
 -- |Fix. Figure out a way to get this from .cabal file
 cabalVersion :: Version
-cabalVersion = Version [1,1,4] []
+cabalVersion = Version [1,1,6] []
 
 -- | This data type is the internal representation of the file @pkg.cabal@.
 -- It contains two kinds of information about the package: information
@@ -182,8 +185,8 @@ emptyPackageDescription
 -- |Get all the module names from the libraries in this package
 libModules :: PackageDescription -> [String]
 libModules PackageDescription{library=lib}
-    = (maybe [] exposedModules lib)
-      ++ (maybe [] (otherModules . libBuildInfo) lib)
+    = maybe [] exposedModules lib
+       ++ maybe [] (otherModules . libBuildInfo) lib
 
 -- |Get all the module names from the exes in this package
 exeModules :: PackageDescription -> [String]
@@ -194,6 +197,11 @@ exeModules PackageDescription{executables=execs}
 hasLibs :: PackageDescription -> Bool
 hasLibs p = maybe False (buildable . libBuildInfo) (library p)
 
+-- |'Maybe' version of 'hasLibs'
+maybeHasLibs :: PackageDescription -> Maybe Library
+maybeHasLibs p =
+   library p >>= (\lib -> toMaybe (buildable (libBuildInfo lib)) lib)
+
 -- Consider refactoring into executable and library versions.
 data BuildInfo = BuildInfo {
         buildable         :: Bool,      -- ^ component is buildable here
@@ -208,6 +216,7 @@ data BuildInfo = BuildInfo {
         extraLibDirs      :: [String],
         includeDirs       :: [FilePath], -- ^directories to find .h files
         includes          :: [FilePath], -- ^ The .h files to be found in includeDirs
+	installIncludes   :: [FilePath], -- ^ .h files to install with the package
         options           :: [(CompilerFlavor,[String])],
         ghcProfOptions       :: [String]
     }
@@ -227,6 +236,7 @@ emptyBuildInfo = BuildInfo {
                       extraLibDirs      = [],
                       includeDirs       = [],
                       includes          = [],
+                      installIncludes   = [],
                       options           = [],
                       ghcProfOptions       = []
                      }
@@ -263,9 +273,8 @@ emptyHookedBuildInfo = (Nothing, [])
 -- |If the package description has a library section, call the given
 --  function with the library build info as argument.
 withLib :: PackageDescription -> a -> (Library -> IO a) -> IO a
-withLib pkg_descr a f = if hasLibs pkg_descr
-                        then f (fromJust (library pkg_descr))
-                        else return a
+withLib pkg_descr a f =
+   maybe (return a) f (maybeHasLibs pkg_descr)
 
 setupMessage :: String -> PackageDescription -> IO ()
 setupMessage msg pkg_descr = 
@@ -316,6 +325,7 @@ unionBuildInfo b1 b2
          extraLibDirs      = combine extraLibDirs,
          includeDirs       = combine includeDirs,
          includes          = combine includes,
+         installIncludes   = combine installIncludes,
          options           = combine options
         }
       where 
@@ -333,6 +343,10 @@ autogenModuleName pkg_descr =
   where fixchar '-' = '_'
         fixchar c   = c
 
+haddockName :: PackageDescription -> FilePath
+haddockName pkg_descr =
+   joinFileExt (pkgName (package pkg_descr)) "haddock"
+
 -- ------------------------------------------------------------
 -- * Parsing & Pretty printing
 -- ------------------------------------------------------------
@@ -355,7 +369,7 @@ basicStanzaFields =
                            (text . showVersion)   parseVersion
                            (pkgVersion . package) (\ver pkg -> pkg{package=(package pkg){pkgVersion=ver}})
  , simpleField "cabal-version"
-                           (text . showVersionRange) parseCabalVersion
+                           (text . showVersionRange) parseVersionRange
                            descCabalVersion       (\v pkg -> pkg{descCabalVersion=v})
  , simpleField "license"
                            (text . show)          parseLicenseQ
@@ -443,6 +457,9 @@ binfoFields =
  , listField   "includes"
                            showFilePath       parseFilePathQ
                            includes           (\paths binfo -> binfo{includes=paths})
+ , listField   "install-includes"
+                           showFilePath       parseFilePathQ
+                           includes           (\paths binfo -> binfo{installIncludes=paths})
  , listField   "include-dirs"
                            showFilePath       parseFilePathQ
                            includeDirs        (\paths binfo -> binfo{includeDirs=paths})
@@ -461,6 +478,8 @@ binfoFields =
                            options            (\path  binfo -> binfo{options=path})
  , optsField   "nhc-options"  NHC
                            options            (\path  binfo -> binfo{options=path})
+ , optsField   "jhc-options"  JHC
+                           options            (\path  binfo -> binfo{options=path})
  ]
 
 
@@ -478,8 +497,9 @@ readAndParseFile parser fpath = do
     ParseFailed e -> do
         let (lineNo, message) = locatedErrorMsg e
         dieWithLocation fpath lineNo message
-    ParseOk x -> return x
-  where
+    ParseOk ws x -> do
+        mapM_ warn ws
+        return x
 
 -- |Parse the given package file.
 readPackageDescription :: FilePath -> IO PackageDescription
@@ -517,7 +537,7 @@ parseDescription inp = do (st:sts) <- splitStanzas inp
         parseExecutableStanza st@((lineNo, "executable",eName):_) =
           case lookupField "main-is" st of
             Just (_,_) -> foldM (parseExecutableField executableStanzaFields) emptyExecutable st
-            Nothing           -> syntaxError lineNo $ "No 'Main-Is' field found for " ++ eName ++ " stanza"
+            Nothing    -> syntaxError lineNo $ "No 'Main-Is' field found for " ++ eName ++ " stanza"
         parseExecutableStanza ((lineNo, f,_):_) = 
           syntaxError lineNo $ "'Executable' stanza starting with field '" ++ f ++ "'"
         parseExecutableStanza _ = error "This shouldn't happen!"
@@ -531,21 +551,15 @@ parseDescription inp = do (st:sts) <- splitStanzas inp
 
         -- ...
         lookupField :: String -> Stanza -> Maybe (LineNo,String)
+        lookupField x sts = lookup x (map (\(n,f,v) -> (f,(n,v))) sts)
 
-        lookupField x ((n,f,v):st)
-          | x == f      = Just (n,v)
-          | otherwise   = lookupField x st
-        lookupField _ [] = Nothing
 
 parseHookedBuildInfo :: String -> ParseResult HookedBuildInfo
 parseHookedBuildInfo inp = do
   stanzas@(mLibStr:exes) <- splitStanzas inp
   mLib <- parseLib mLibStr
-  case mLib of
-    Nothing      -> do biExes <- mapM parseExe stanzas
-                       return (Nothing, biExes)
-    lib@(Just _) -> do biExes <- mapM parseExe exes
-                       return (lib, biExes)
+  biExes <- mapM parseExe (maybe stanzas (const exes) mLib)
+  return (mLib, biExes)
   where
     parseLib :: Stanza -> ParseResult (Maybe BuildInfo)
     parseLib (bi@((_, inFieldName, _):_))
@@ -565,8 +579,11 @@ parseBInfoField :: [StanzaField a] -> a -> (LineNo, String, String) -> ParseResu
 parseBInfoField ((StanzaField name _ set):fields) binfo (lineNo, f, val)
           | name == f = set lineNo val binfo
           | otherwise = parseBInfoField fields binfo (lineNo, f, val)
-parseBInfoField [] _ (lineNo, f, _) =
-          syntaxError lineNo $ "Unknown field '" ++ f ++ "'"
+-- ignore "x-" extension fields without a warning
+parseBInfoField [] binfo (lineNo, 'x':'-':f, _) = return binfo
+parseBInfoField [] binfo (lineNo, f, _) = do
+          warning $ "Unknown field '" ++ f ++ "'"
+          return binfo
 
 -- --------------------------------------------
 -- ** Pretty printing
@@ -656,6 +673,9 @@ errorOut warnings errors = do
     mapM (hPutStrLn stderr . ((pname ++ ": Error: ") ++)) errors
     exitWith (ExitFailure 1)
 
+toMaybe :: Bool -> a -> Maybe a
+toMaybe b x = if b then Just x else Nothing
+
 checkMissingFields :: PackageDescription -> [Maybe String]
 checkMissingFields pkg_descr = 
     [missingField (pkgName . package)    reqNameName
@@ -665,31 +685,23 @@ checkMissingFields pkg_descr =
                        -> String -- Name of field
                        -> Maybe String -- error message
           missingField f n
-              = if null (f pkg_descr)
-                 then Just $ "Missing field: " ++ n
-                 else Nothing
+              = toMaybe (null (f pkg_descr)) ("Missing field: " ++ n)
 
 sanityCheckLib :: Maybe Library -> Maybe String
-sanityCheckLib Nothing  = Nothing
-sanityCheckLib (Just l)
-    = if null $ exposedModules l
-      then Just "Non-empty library, but empty exposed modules list. Cabal may not build this library correctly"
-      else Nothing
+sanityCheckLib ml =
+   ml >>= (\l ->
+      toMaybe (null $ exposedModules l)
+              ("Non-empty library, but empty exposed modules list. " ++
+               "Cabal may not build this library correctly"))
 
 checkSanity :: Bool -> String -> Maybe String
-checkSanity False _  = Nothing
-checkSanity True  s  = Just s
+checkSanity = toMaybe
 
 hasMods :: PackageDescription -> Bool
-hasMods pkg_descr
-    | isNothing (library pkg_descr) = null $ executables pkg_descr
-    | otherwise = null (executables pkg_descr)
-                   && null (exposedModules (fromJust (library pkg_descr)))
+hasMods pkg_descr =
+   null (executables pkg_descr) &&
+      maybe True (null . exposedModules) (library pkg_descr)
 
-parseCabalVersion = do v <- parseVersionRange
-                       if (cabalVersion `withinRange` v)
-                          then return v
-                          else error ("This package requires Cabal verion: " ++ (showVersionRange v) ++ ".")
 
 -- ------------------------------------------------------------
 -- * Testing
@@ -730,9 +742,11 @@ testPkgDesc = unlines [
         "Extra-Lib-Dirs: \"/usr/local/libs\"",
         "Include-Dirs: your/slightest, look/will",
         "Includes: /easily/unclose, /me, \"funky, path\\\\name\"",
+        "Install-Includes: /easily/unclose, /me, \"funky, path\\\\name\"",
         "GHC-Options: -fTH -fglasgow-exts",
         "Hugs-Options: +TH",
         "Nhc-Options: ",
+        "Jhc-Options: ",
         "",
         "-- Next is an executable",
         "Executable: somescript",
@@ -742,7 +756,8 @@ testPkgDesc = unlines [
         "Extensions: OverlappingInstances",
         "GHC-Options: ",
         "Hugs-Options: ",
-        "Nhc-Options: "
+        "Nhc-Options: ",
+        "Jhc-Options: "
         ]
 
 testPkgDescAnswer :: PackageDescription
@@ -788,9 +803,10 @@ testPkgDescAnswer =
                            extraLibDirs = ["/usr/local/libs"],
                            includeDirs = ["your/slightest", "look/will"],
                            includes = ["/easily/unclose", "/me", "funky, path\\name"],
+                           installIncludes = ["/easily/unclose", "/me", "funky, path\\name"],
                            -- Note reversed order:
                            ghcProfOptions = [],
-                           options = [(NHC, []), (Hugs,["+TH"]), (GHC,["-fTH","-fglasgow-exts"])]}
+                           options = [(JHC,[]),(NHC, []), (Hugs,["+TH"]), (GHC,["-fTH","-fglasgow-exts"])]}
                     },
                     executables = [Executable "somescript" 
                        "SomeFile.hs" (
@@ -798,7 +814,7 @@ testPkgDescAnswer =
                         otherModules=["Foo1","Util","Main"],
                         hsSourceDirs = ["scripts"],
                         extensions = [OverlappingInstances],
-                        options = [(NHC,[]),(Hugs,[]),(GHC,[])]
+                        options = [(JHC,[]),(NHC,[]),(Hugs,[]),(GHC,[])]
                       })]
 }
 
@@ -836,10 +852,10 @@ hunitTests = [
              TestLabel "Package description pretty" $ TestCase $ 
                 case parseDescription testPkgDesc of
                  ParseFailed _ -> assertBool "can't parse description" False
-                 ParseOk d -> case parseDescription $ showPackageDescription d of
+                 ParseOk _ d -> case parseDescription $ showPackageDescription d of
                                 ParseFailed _ ->
                                     assertBool "can't parse description after pretty print!" False
-                                ParseOk d' -> 
+                                ParseOk _ d' -> 
                                     assertBool ("parse . show . parse not identity."
                                                 ++"   Incorrect fields:"
                                                 ++ (show $ comparePackageDescriptions d d'))
@@ -863,17 +879,16 @@ comparePackageDescriptions p1 p2
                   -> Maybe String -- 
             myCmp f er = let e1 = f p1
                              e2 = f p2
-                          in if e1 == e2
-                              then Nothing
-                              else Just (er ++ " Expected: " ++ show e1
-                                            ++ " Got: " ++ show e2)
+                          in toMaybe (e1 /= e2)
+                                     (er ++ " Expected: " ++ show e1
+                                              ++ " Got: " ++ show e2)
 
 -- |Assert that the 2nd value parses correctly and matches the first value
 assertParseOk :: (Eq val) => String -> val -> ParseResult val -> Assertion
 assertParseOk mes expected actual
     =  assertBool mes
            (case actual of
-             ParseOk v -> v == expected
+             ParseOk _ v -> v == expected
              _         -> False)
 
 test :: IO Counts
