@@ -1,3 +1,4 @@
+{-# OPTIONS -cpp #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Distribution.Setup
@@ -44,8 +45,13 @@ module Distribution.Setup (--parseArgs,
                            module Distribution.Compiler,
                            Action(..),
                            ConfigFlags(..), emptyConfigFlags, configureArgs,
-                           CopyFlags, CopyDest(..), InstallFlags, RegisterFlags, SDistFlags,
-                           InstallUserFlag(..),
+                           CopyFlags(..), CopyDest(..), 
+			   InstallFlags(..), emptyInstallFlags,
+                           HaddockFlags(..), emptyHaddockFlags,
+                           BuildFlags(..), CleanFlags(..), PFEFlags(..),
+                           RegisterFlags(..), emptyRegisterFlags,
+			   SDistFlags(..),
+                           MaybeUserFlag(..), userOverride,
 			   --optionHelpString,
 #ifdef DEBUG
                            hunitTests,
@@ -66,12 +72,10 @@ import HUnit (Test(..))
 
 import Distribution.Compiler (CompilerFlavor(..), Compiler(..))
 import Distribution.Simple.Utils (die)
-import Distribution.Program(ProgramLocation(..), ProgramConfiguration(..),
-                            Program(..),
-                            defaultProgramConfiguration, userSpecifyPath,
-                            userSpecifyArgs, haddockProgram)
+import Distribution.Program(ProgramConfiguration(..),
+                            userSpecifyPath, userSpecifyArgs)
 import Data.List(find)
-import Data.FiniteMap(keysFM)
+import Distribution.Compat.Map (keys)
 import Distribution.GetOpt
 import Distribution.Compat.FilePath (platformPath)
 import System.Exit
@@ -80,24 +84,28 @@ import System.Environment
 -- type CommandLineOpts = (Action,
 --                         [String]) -- The un-parsed remainder
 
-data Action = ConfigCmd ConfigFlags       -- config
-            | BuildCmd                    -- build
-            | CleanCmd                    -- clean
-            | CopyCmd CopyDest            -- copy (--destdir flag)
-            | HaddockCmd                  -- haddock
-            | ProgramaticaCmd             -- pfesetup
-            | InstallCmd InstallUserFlag -- install (install-prefix) (--user flag)
-            | SDistCmd                    -- sdist
-            | TestCmd                     -- test
-            | RegisterCmd   Bool Bool     -- register (--user flag, --gen-script)
-            | UnregisterCmd Bool Bool     -- unregister (--user flag, --gen-script)
-	    | HelpCmd			  -- help
+data Action = ConfigCmd ConfigFlags   -- config
+            | BuildCmd                -- build
+            | CleanCmd                -- clean
+            | CopyCmd CopyDest        -- copy (--destdir flag)
+            | HaddockCmd              -- haddock
+            | ProgramaticaCmd         -- pfesetup
+            | InstallCmd              -- install (install-prefix)
+            | SDistCmd                -- sdist
+            | TestCmd                 -- test
+            | RegisterCmd    	      -- register
+            | UnregisterCmd           -- unregister
+	    | HelpCmd		      -- help
 --            | NoCmd -- error case, help case.
 --             | TestCmd 1.0?
 --             | BDist -- 1.0
 --            | CleanCmd                 -- clean
 --            | NoCmd -- error case?
 
+-- ------------------------------------------------------------
+-- * Flag-related types
+-- ------------------------------------------------------------
+
 -- | Flags to @configure@ command
 data ConfigFlags = ConfigFlags {
         configPrograms :: ProgramConfiguration, -- ^All programs that cabal may run
@@ -110,6 +118,7 @@ data ConfigFlags = ConfigFlags {
         configC2hs     :: Maybe FilePath, -- ^C2hs path
         configCpphs    :: Maybe FilePath, -- ^Cpphs path
         configGreencard:: Maybe FilePath, -- ^GreenCard path
+        configVanillaLib  :: Bool,        -- ^Enable vanilla library
         configProfLib  :: Bool,           -- ^Enable profiling in the library
         configProfExe  :: Bool,           -- ^Enable profiling in the executables.
         configPrefix   :: Maybe FilePath,
@@ -129,7 +138,8 @@ data ConfigFlags = ConfigFlags {
 
         configVerbose  :: Int,            -- ^verbosity level
 	configUser     :: Bool,		  -- ^--user flag?
-	configGHCiLib  :: Bool            -- ^Enable compiling library for GHCi
+	configGHCiLib  :: Bool,           -- ^Enable compiling library for GHCi
+	configSplitObjs :: Bool		  -- ^Enable -split-objs with GHC
     }
 
 emptyConfigFlags :: ProgramConfiguration -> ConfigFlags
@@ -143,6 +153,7 @@ emptyConfigFlags progConf = ConfigFlags {
         configAlex     = Nothing,
         configHsc2hs   = Nothing,
         configC2hs     = Nothing,
+        configVanillaLib  = True,
         configProfLib  = False,
         configProfExe  = False,
         configCpphs    = Nothing,
@@ -156,15 +167,125 @@ emptyConfigFlags progConf = ConfigFlags {
 	configDataSubDir = Nothing,
         configVerbose  = 0,
 	configUser     = False,
-	configGHCiLib  = True
+	configGHCiLib  = True,
+	configSplitObjs = False -- takes longer, so turn off by default
     }
 
+-- | Flags to @copy@: (destdir, copy-prefix (backwards compat), verbose)
+data CopyFlags = CopyFlags {copyDest      :: CopyDest
+                           ,copyVerbose :: Int}
+
+data CopyDest
+  = NoCopyDest
+  | CopyTo FilePath
+  | CopyPrefix FilePath		-- DEPRECATED
+  deriving (Eq, Show)
+
+data MaybeUserFlag  = MaybeUserNone   -- ^no --user OR --global flag.
+                    | MaybeUserUser   -- ^--user flag
+                    | MaybeUserGlobal -- ^--global flag
+
+-- |A 'MaybeUserFlag' overrides the default --user setting
+userOverride :: MaybeUserFlag -> Bool -> Bool
+MaybeUserUser   `userOverride` _ = True
+MaybeUserGlobal `userOverride` _ = False
+_               `userOverride` r = r
+
+-- | Flags to @install@: (user package, verbose)
+data InstallFlags = InstallFlags {installUserFlags::MaybeUserFlag
+                                 ,installVerbose :: Int}
+
+emptyInstallFlags :: InstallFlags
+emptyInstallFlags = InstallFlags{ installUserFlags=MaybeUserNone,
+				  installVerbose=0 }
+
+-- | Flags to @sdist@: (snapshot, verbose)
+data SDistFlags = SDistFlags {sDistSnapshot::Bool
+                             ,sDistVerbose:: Int}
+
+-- | Flags to @register@ and @unregister@: (user package, gen-script, 
+-- in-place, verbose)
+data RegisterFlags = RegisterFlags {regUser::MaybeUserFlag
+                                   ,regGenScript::Bool
+				   ,regInPlace::Bool
+				   ,regWithHcPkg::Maybe FilePath
+                                   ,regVerbose::Int}
+
+
+emptyRegisterFlags :: RegisterFlags
+emptyRegisterFlags = RegisterFlags { regUser=MaybeUserNone,
+				     regGenScript=False,
+				     regInPlace=False,
+				     regWithHcPkg=Nothing,
+				     regVerbose=0 }
+
+data HaddockFlags = HaddockFlags {haddockHoogle :: Bool
+                                 ,haddockVerbose :: Int}
+
+emptyHaddockFlags :: HaddockFlags
+emptyHaddockFlags = HaddockFlags {haddockHoogle = False, haddockVerbose = 0}
+
+-- Following only have verbose flags, but for consistency and
+-- extensibility we make them into a type.
+data BuildFlags   = BuildFlags   {buildVerbose   :: Int}
+data CleanFlags   = CleanFlags   {cleanVerbose   :: Int}
+data PFEFlags     = PFEFlags     {pfeVerbose     :: Int}
+
+-- |Most of these flags are for Configure, but InstPrefix is for Copy.
+data Flag a = GhcFlag | NhcFlag | HugsFlag | JhcFlag
+          | WithCompiler FilePath | WithHcPkg FilePath
+          | WithHappy FilePath | WithAlex FilePath
+          | WithHsc2hs FilePath | WithC2hs FilePath | WithCpphs FilePath
+          | WithGreencard FilePath
+          | WithVanillaLib | WithoutVanillaLib
+          | WithProfLib | WithoutProfLib
+          | WithProfExe | WithoutProfExe
+	  | WithGHCiLib | WithoutGHCiLib
+	  | WithSplitObjs | WithoutSplitObjs
+
+	  | Prefix FilePath
+	  | BinDir FilePath
+	  | LibDir FilePath
+	  | LibSubDir FilePath
+	  | LibExecDir FilePath
+	  | DataDir FilePath
+	  | DataSubDir FilePath
+
+          | ProgramArgs String String   -- program name, arguments
+          | WithProgram String FilePath -- program name, location
+
+          -- For install, register, and unregister:
+          | UserFlag | GlobalFlag
+          -- for register & unregister
+          | GenScriptFlag
+	  | InPlaceFlag
+          -- For copy:
+          | InstPrefix FilePath
+	  | DestDir FilePath
+          -- For sdist:
+          | Snapshot
+          -- For haddock:
+          | HaddockHoogle
+          -- For everyone:
+          | HelpFlag
+          | Verbose Int
+--          | Version?
+          | Lift a
+            deriving (Show, Eq)
+
+
+-- ------------------------------------------------------------
+-- * Mostly parsing functions
+-- ------------------------------------------------------------
+
 defaultCompilerFlavor :: Maybe CompilerFlavor
 defaultCompilerFlavor =
 #if defined(__GLASGOW_HASKELL__)
    Just GHC
 #elif defined(__NHC__)
    Just NHC
+#elif defined(__JHC__)
+   Just JHC
 #elif defined(__HUGS__)
    Just Hugs
 #else
@@ -193,45 +314,10 @@ configureArgs flags
 
         showHC GHC = "ghc"
         showHC NHC = "nhc98"
+        showHC JHC = "jhc"
         showHC Hugs = "hugs"
         showHC c    = "unknown compiler: " ++ (show c)
 
--- |Most of these flags are for Configure, but InstPrefix is for Copy.
-data Flag a = GhcFlag | NhcFlag | HugsFlag
-          | WithCompiler FilePath | WithHcPkg FilePath
-          | WithHappy FilePath | WithAlex FilePath
-          | WithHsc2hs FilePath | WithC2hs FilePath | WithCpphs FilePath
-          | WithGreencard FilePath
-          | WithProfLib | WithoutProfLib
-          | WithProfExe | WithoutProfExe
-	  | WithGHCiLib | WithoutGHCiLib
-
-	  | Prefix FilePath
-	  | BinDir FilePath
-	  | LibDir FilePath
-	  | LibSubDir FilePath
-	  | LibExecDir FilePath
-	  | DataDir FilePath
-	  | DataSubDir FilePath
-
-          | ProgramArgs String String   -- program name, arguments
-          | WithProgram String FilePath -- program name, location
-
-          -- For install, register, and unregister:
-          | UserFlag | GlobalFlag
-          -- for register & unregister
-          | GenScriptFlag
-          -- For copy:
-          | InstPrefix FilePath
-	  | DestDir FilePath
-          -- For sdist:
-          | Snapshot
-          -- For everyone:
-          | HelpFlag
-          | Verbose Int
---          | Version?
-          | Lift a
-            deriving (Show, Eq)
 
 cmd_help :: OptDescr (Flag a)
 cmd_help = Option "h?" ["help"] (NoArg HelpFlag) "Show this help text"
@@ -241,6 +327,10 @@ cmd_verbose = Option "v" ["verbose"] (OptArg verboseFlag "n") "Control verbosity
   where
     verboseFlag mb_s = Verbose (maybe 3 read mb_s)
 
+cmd_with_hc_pkg :: OptDescr (Flag a)
+cmd_with_hc_pkg = Option "" ["with-hc-pkg"] (reqPathArg WithHcPkg)
+               "give the path to the package tool"
+
 -- Do we have any other interesting global flags?
 globalOptions :: [OptDescr (Flag a)]
 globalOptions = [
@@ -305,9 +395,9 @@ parseGlobalArgs progConf args =
     (flags, _, _, []) | hasHelpFlag flags -> do
       (printGlobalHelp progConf)
       exitWith ExitSuccess
-    (_, cname:cargs, _, []) -> do
+    (_, cname:cargs, extra_args, []) -> do
       case lookupCommand cname (commandList progConf) of
-        Just cmd -> return (cmdAction cmd,cargs)
+        Just cmd -> return (cmdAction cmd, extra_args ++ cargs)
         Nothing  -> die $ "Unrecognised command: " ++ cname ++ " (try --help)"
     (_, [], _, [])  -> die $ "No command given (try --help)"
     (_, _, _, errs) -> putErrors errs
@@ -320,11 +410,11 @@ configureCmd progConf = Cmd {
         cmdOptions     = [cmd_help, cmd_verbose,
            Option "g" ["ghc"] (NoArg GhcFlag) "compile with GHC",
            Option "n" ["nhc"] (NoArg NhcFlag) "compile with NHC",
+           Option "" ["jhc"]  (NoArg JhcFlag) "compile with JHC",
            Option "" ["hugs"] (NoArg HugsFlag) "compile with hugs",
            Option "w" ["with-compiler"] (reqPathArg WithCompiler)
                "give the path to a particular compiler",
-           Option "" ["with-hc-pkg"] (reqPathArg WithHcPkg)
-               "give the path to the package tool",
+	   cmd_with_hc_pkg,
            Option "" ["prefix"] (reqDirArg Prefix)
                "bake this prefix in preparation of installation",
 	   Option "" ["bindir"] (reqDirArg BinDir)
@@ -351,6 +441,10 @@ configureCmd progConf = Cmd {
                "give the path to cpphs",
            Option "" ["with-greencard"] (reqPathArg WithGreencard)
                "give the path to greencard",
+           Option "" ["enable-library-vanilla"] (NoArg WithVanillaLib)
+               "Enable vanilla libraries",
+           Option "" ["disable-library-vanilla"] (NoArg WithoutVanillaLib)
+               "Disable vanilla libraries",
            Option "p" ["enable-library-profiling"] (NoArg WithProfLib)
                "Enable library profiling",
            Option "" ["disable-library-profiling"] (NoArg WithoutProfLib)
@@ -363,6 +457,10 @@ configureCmd progConf = Cmd {
                "compile library for use with GHCi",
 	   Option "" ["disable-library-for-ghci"] (NoArg WithoutGHCiLib)
                "do not compile libraries for GHCi",
+	   Option "" ["enable-split-objs"] (NoArg WithSplitObjs)
+	       "split library into smaller objects to reduce binary sizes (GHC 6.6+)",
+	   Option "" ["disable-split-objs"] (NoArg WithoutSplitObjs)
+	       "split library into smaller objects to reduce binary sizes (GHC 6.6+)",
            Option "" ["user"] (NoArg UserFlag)
                "allow dependencies to be satisfied from the user package database. also implies install --user",
            Option "" ["global"] (NoArg GlobalFlag)
@@ -381,12 +479,12 @@ configureCmd progConf = Cmd {
         }
 
 programArgsOptions :: ProgramConfiguration -> [OptDescr (Flag a)]
-programArgsOptions (ProgramConfiguration conf) = map f (keysFM conf)
+programArgsOptions (ProgramConfiguration conf) = map f (keys conf)
     where f name = Option "" [name ++ "-args"] (reqPathArg (ProgramArgs name))
                    ("give the args to " ++ name)
 
 withProgramOptions :: ProgramConfiguration -> [OptDescr (Flag a)]
-withProgramOptions (ProgramConfiguration conf) = map f (keysFM conf)
+withProgramOptions (ProgramConfiguration conf) = map f (keys conf)
     where f name = Option "" ["with-" ++ name] (reqPathArg (WithProgram name))
                    ("give the path to " ++ name)
 
@@ -401,6 +499,7 @@ parseConfigureArgs :: ProgramConfiguration -> ConfigFlags -> [String] -> [OptDes
 parseConfigureArgs progConf = parseArgs (configureCmd progConf) updateCfg
   where updateCfg t GhcFlag              = t { configHcFlavor = Just GHC }
         updateCfg t NhcFlag              = t { configHcFlavor = Just NHC }
+        updateCfg t JhcFlag              = t { configHcFlavor = Just JHC }
         updateCfg t HugsFlag             = t { configHcFlavor = Just Hugs }
         updateCfg t (WithCompiler path)  = t { configHcPath   = Just path }
         updateCfg t (WithHcPkg path)     = t { configHcPkg    = Just path }
@@ -416,6 +515,8 @@ parseConfigureArgs progConf = parseArgs (configureCmd progConf) updateCfg
         updateCfg t (WithProgram name path) = t { configPrograms = (userSpecifyPath
                                                                  name
                                                                  path (configPrograms t))}
+        updateCfg t WithVanillaLib       = t { configVanillaLib  = True }
+        updateCfg t WithoutVanillaLib    = t { configVanillaLib  = False, configGHCiLib = False }
         updateCfg t WithProfLib          = t { configProfLib  = True }
         updateCfg t WithoutProfLib       = t { configProfLib  = False }
         updateCfg t WithProfExe          = t { configProfExe  = True }
@@ -432,6 +533,8 @@ parseConfigureArgs progConf = parseArgs (configureCmd progConf) updateCfg
         updateCfg t (Verbose n)          = t { configVerbose  = n }
         updateCfg t UserFlag             = t { configUser     = True }
         updateCfg t GlobalFlag           = t { configUser     = False }
+	updateCfg t WithSplitObjs	 = t { configSplitObjs = True }
+	updateCfg t WithoutSplitObjs	 = t { configSplitObjs = False }
         updateCfg t (Lift _)             = t
         updateCfg t _                    = error $ "Unexpected flag!"
 
@@ -444,18 +547,26 @@ buildCmd = Cmd {
         cmdAction      = BuildCmd
         }
 
-parseBuildArgs :: [String] -> [OptDescr a] -> IO (Int, [a], [String])
-parseBuildArgs = parseNoArgs buildCmd
+parseBuildArgs :: [String] -> [OptDescr a] -> IO (BuildFlags, [a], [String])
+parseBuildArgs = parseNoArgs buildCmd BuildFlags
 
 haddockCmd :: Cmd a
 haddockCmd = Cmd {
         cmdName        = "haddock",
         cmdHelp        = "Generate Haddock HTML code from Exposed-Modules.",
         cmdDescription = "Requires cpphs and haddock.",
-        cmdOptions     = [cmd_help, cmd_verbose],
+        cmdOptions     = [cmd_help, cmd_verbose,
+                          Option "" ["hoogle"] (NoArg HaddockHoogle) "Generate a hoogle database"],
         cmdAction      = HaddockCmd
         }
 
+parseHaddockArgs :: HaddockFlags -> [String] -> [OptDescr a] -> IO (HaddockFlags, [a], [String])
+parseHaddockArgs  = parseArgs haddockCmd updateCfg
+  where updateCfg (HaddockFlags hoogle verbose) fl = case fl of
+            HaddockHoogle -> HaddockFlags True verbose
+            Verbose n     -> HaddockFlags hoogle n
+            _             -> error "Unexpected flag!"
+
 programaticaCmd :: Cmd a
 programaticaCmd = Cmd {
         cmdName        = "pfe",
@@ -465,11 +576,8 @@ programaticaCmd = Cmd {
         cmdAction      = ProgramaticaCmd
         }
 
-parseProgramaticaArgs :: [String] -> [OptDescr a] -> IO (Int, [a], [String])
-parseProgramaticaArgs  = parseNoArgs programaticaCmd
-
-parseHaddockArgs :: [String] -> [OptDescr a] -> IO (Int, [a], [String])
-parseHaddockArgs  = parseNoArgs haddockCmd
+parseProgramaticaArgs :: [String] -> [OptDescr a] -> IO (PFEFlags, [a], [String])
+parseProgramaticaArgs  = parseNoArgs programaticaCmd PFEFlags
 
 cleanCmd :: Cmd a
 cleanCmd = Cmd {
@@ -480,8 +588,8 @@ cleanCmd = Cmd {
         cmdAction      = CleanCmd
         }
 
-parseCleanArgs :: [String] -> [OptDescr a] -> IO (Int, [a], [String])
-parseCleanArgs = parseNoArgs cleanCmd
+parseCleanArgs :: [String] -> [OptDescr a] -> IO (CleanFlags, [a], [String])
+parseCleanArgs = parseNoArgs cleanCmd CleanFlags
 
 installCmd :: Cmd a
 installCmd = Cmd {
@@ -496,7 +604,7 @@ installCmd = Cmd {
            Option "" ["global"] (NoArg GlobalFlag)
                "(default; override with configure) upon registration, register this package in the system-wide package database"
            ],
-        cmdAction      = InstallCmd InstallUserNone
+        cmdAction      = InstallCmd
         }
 
 copyCmd :: Cmd a
@@ -513,39 +621,24 @@ copyCmd = Cmd {
         cmdAction      = CopyCmd NoCopyDest
         }
 
--- | Flags to @copy@: (destdir, copy-prefix (backwards compat), verbose)
-type CopyFlags = (CopyDest,Int)
-
-data CopyDest
-  = NoCopyDest
-  | CopyTo FilePath
-  | CopyPrefix FilePath		-- DEPRECATED
-  deriving (Eq, Show)
-
-data InstallUserFlag = InstallUserNone   -- ^no --user OR --global flag.
-                      | InstallUserUser   -- ^--user flag
-                      | InstallUserGlobal -- ^--global flag
-
 parseCopyArgs :: CopyFlags -> [String] -> [OptDescr a] ->
                     IO (CopyFlags, [a], [String])
 parseCopyArgs = parseArgs copyCmd updateCfg
-  where updateCfg (copydest,verbose) fl = case fl of
-            InstPrefix path -> (CopyPrefix path, verbose)
-	    DestDir path    -> (CopyTo path,verbose)
-            Verbose n       -> (copydest,n)
+  where updateCfg (CopyFlags copydest verbose) fl = case fl of
+            InstPrefix path -> (CopyFlags (CopyPrefix path) verbose)
+	    DestDir path    -> (CopyFlags (CopyTo path) verbose)
+            Verbose n       -> (CopyFlags copydest n)
             _               -> error $ "Unexpected flag!"
 
--- | Flags to @install@: (user package, verbose)
-type InstallFlags = (InstallUserFlag,Int)
 
 parseInstallArgs :: InstallFlags -> [String] -> [OptDescr a] ->
                     IO (InstallFlags, [a], [String])
 parseInstallArgs = parseArgs installCmd updateCfg
-  where updateCfg (uFlag,verbose) fl = case fl of
+  where updateCfg (InstallFlags uFlag verbose) fl = case fl of
             InstPrefix _ -> error "--install-prefix is obsolete. Use copy command instead."
-            UserFlag     -> (InstallUserUser,  verbose)
-            GlobalFlag   -> (InstallUserGlobal, verbose)
-            Verbose n    -> (uFlag, n)
+            UserFlag     -> (InstallFlags MaybeUserUser  verbose)
+            GlobalFlag   -> (InstallFlags MaybeUserGlobal verbose)
+            Verbose n    -> (InstallFlags uFlag n)
             _            -> error $ "Unexpected flag!"
 
 sdistCmd :: Cmd a
@@ -560,14 +653,11 @@ sdistCmd = Cmd {
         cmdAction      = SDistCmd
         }
 
--- | Flags to @sdist@: (snapshot, verbose)
-type SDistFlags = (Bool, Int)
-
 parseSDistArgs :: [String] -> [OptDescr a] -> IO (SDistFlags, [a], [String])
-parseSDistArgs = parseArgs sdistCmd updateCfg (False, 0)
-  where updateCfg (snapshot, verbose) fl = case fl of
-            Snapshot        -> (True,     verbose)
-            Verbose n       -> (snapshot, n)
+parseSDistArgs = parseArgs sdistCmd updateCfg (SDistFlags False 0)
+  where updateCfg (SDistFlags snapshot verbose) fl = case fl of
+            Snapshot        -> (SDistFlags True verbose)
+            Verbose n       -> (SDistFlags snapshot n)
             _               -> error $ "Unexpected flag!"
 
 testCmd :: Cmd a
@@ -580,7 +670,7 @@ testCmd = Cmd {
         }
 
 parseTestArgs :: [String] -> [OptDescr a] -> IO (Int, [a], [String])
-parseTestArgs = parseNoArgs testCmd
+parseTestArgs = parseNoArgs testCmd id
 
 registerCmd :: Cmd a
 registerCmd = Cmd {
@@ -592,23 +682,25 @@ registerCmd = Cmd {
                "upon registration, register this package in the user's local package database",
            Option "" ["global"] (NoArg GlobalFlag)
                "(default) upon registration, register this package in the system-wide package database",
+           Option "" ["inplace"] (NoArg InPlaceFlag)
+               "register the package in the build location, so it can be used without being installed",
            Option "" ["gen-script"] (NoArg GenScriptFlag)
-               "Instead of performing the register command, generate a script to register later"
+               "Instead of performing the register command, generate a script to register later",
+	   cmd_with_hc_pkg
            ],
-        cmdAction      = RegisterCmd False False
+        cmdAction      = RegisterCmd
         }
 
--- | Flags to @register@ and @unregister@: (user package, gen-script, verbose)
-type RegisterFlags = (Bool, Bool, Int)
-
 parseRegisterArgs :: RegisterFlags -> [String] -> [OptDescr a] ->
                      IO (RegisterFlags, [a], [String])
 parseRegisterArgs = parseArgs registerCmd updateCfg
-  where updateCfg (uFlag, genScriptFlag, verbose) fl = case fl of
-            UserFlag        -> (True,  genScriptFlag, verbose)
-            GlobalFlag      -> (False, genScriptFlag, verbose)
-            Verbose n       -> (uFlag, genScriptFlag, n)
-            GenScriptFlag   -> (uFlag, True,          verbose)
+  where updateCfg reg fl = case fl of
+            UserFlag        -> reg { regUser=MaybeUserUser }
+            GlobalFlag      -> reg { regUser=MaybeUserGlobal }
+            Verbose n       -> reg { regVerbose=n }
+            GenScriptFlag   -> reg { regGenScript=True }
+	    InPlaceFlag     -> reg { regInPlace=True }
+	    WithHcPkg f	    -> reg { regWithHcPkg=Just f }
             _               -> error $ "Unexpected flag!"
 
 unregisterCmd :: Cmd a
@@ -625,7 +717,7 @@ unregisterCmd = Cmd {
                "Instead of performing the unregister command, generate a script to unregister later"
 
            ],
-        cmdAction      = UnregisterCmd False False
+        cmdAction      = UnregisterCmd
         }
 
 parseUnregisterArgs :: RegisterFlags -> [String] -> [OptDescr a] ->
@@ -635,10 +727,12 @@ parseUnregisterArgs = parseRegisterArgs
 -- |Helper function for commands with no arguments except for verbose
 -- and help.
 
-parseNoArgs :: (Cmd a) -> [String] -> [OptDescr a] -> IO (Int, [a], [String])
-parseNoArgs cmd = parseArgs cmd updateCfg 0
+parseNoArgs :: (Cmd a)
+            -> (Int -> b) -- Constructor to make this type.
+            -> [String] -> [OptDescr a]-> IO (b, [a], [String])
+parseNoArgs cmd c = parseArgs cmd updateCfg (c 0)
   where
-    updateCfg _ (Verbose n) = n
+    updateCfg _ (Verbose n) = c n
     updateCfg _ _           = error "Unexpected flag!"
 
 -- |Helper function for commands with more options.
