@@ -1,4 +1,4 @@
-{-# OPTIONS -fffi #-}
+{-# OPTIONS -cpp -fffi #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Distribution.Simple.Utils
@@ -49,15 +49,17 @@ module Distribution.Simple.Utils (
         rawSystemVerbose,
 	rawSystemExit,
         maybeExit,
+        xargs,
         matchesDescFile,
 	rawSystemPathExit,
         smartCopySources,
         copyFileVerbose,
+        copyDirectoryRecursiveVerbose,
         moduleToFilePath,
         mkLibName,
         mkProfLibName,
-	mkGHCiLibName,
         currentDir,
+	dirOf,
         dotToSep,
 	withTempFile,
 	findFile,
@@ -65,6 +67,9 @@ module Distribution.Simple.Utils (
         findPackageDesc,
 	defaultHookedPackageDesc,
 	findHookedPackageDesc,
+        distPref,
+        haddockPref,
+        srcPref,
 #ifdef DEBUG
         hunitTests
 #endif
@@ -81,8 +86,8 @@ module Distribution.Simple.Utils (
 import Distribution.Compat.RawSystem (rawSystem)
 import Distribution.Compat.Exception (finally)
 
-import Control.Monad(when, filterM)
-import Data.List (nub)
+import Control.Monad(when, filterM, unless)
+import Data.List (nub, unfoldr)
 import System.Environment (getProgName)
 import System.IO (hPutStrLn, stderr, hFlush, stdout)
 import System.IO.Error
@@ -92,13 +97,15 @@ import System.Posix.Internals (c_getpid)
 #endif
 
 import Distribution.Compat.FilePath
-	(splitFileName, splitFileExt, joinFileName, joinFileExt,
-	pathSeparator)
+	(splitFileName, splitFileExt, joinFileName, joinFileExt, joinPaths,
+	pathSeparator,splitFilePath)
 import System.Directory (getDirectoryContents, getCurrentDirectory
-			, doesFileExist, removeFile, getPermissions
+			, doesDirectoryExist, doesFileExist, removeFile, getPermissions
 			, Permissions(executable))
 
-import Distribution.Compat.Directory (copyFile,findExecutable,createDirectoryIfMissing)
+import Distribution.Compat.Directory
+           (copyFile, findExecutable, createDirectoryIfMissing,
+            getDirectoryContentsWithoutSpecial)
 
 #ifdef DEBUG
 import HUnit ((~:), (~=?), Test(..), assertEqual)
@@ -163,6 +170,36 @@ rawSystemPathExit :: Int -> String -> [String] -> IO ()
 rawSystemPathExit verbose prog args = do
   maybeExit $ rawSystemPath verbose prog args
 
+-- | Like the unix xargs program. Useful for when we've got very long command
+-- lines that might overflow an OS limit on command line length and so you
+-- need to invoke a command multiple times to get all the args in.
+--
+-- Use it with either of the rawSystem variants above. For example:
+-- 
+-- > xargs (32*1024) (rawSystemPath verbose) prog fixedArgs bigArgs
+--
+xargs :: Int -> (FilePath -> [String] -> IO ExitCode)
+      -> FilePath -> [String] -> [String] -> IO ExitCode
+xargs maxSize rawSystem prog fixedArgs bigArgs =
+  let fixedArgSize = sum (map length fixedArgs) + length fixedArgs
+      chunkSize = maxSize - fixedArgSize
+      loop [] = return ExitSuccess
+      loop (args:remainingArgs) = do
+        status <- rawSystem prog (fixedArgs ++ args)
+        case status of
+          ExitSuccess -> loop remainingArgs
+          _           -> return status
+   in loop (chunks chunkSize bigArgs)
+
+  where chunks len = unfoldr $ \s ->
+          if null s then Nothing
+                    else Just (chunk [] len s)
+
+        chunk acc len []     = (reverse acc,[])
+        chunk acc len (s:ss)
+          | len' < len = chunk (s:acc) (len-len'-1) ss
+          | otherwise  = (reverse acc, s:ss)
+          where len' = length s
 
 -- ------------------------------------------------------------
 -- * File Utilities
@@ -266,11 +303,38 @@ copyFileVerbose verbose src dest = do
     putStrLn ("copy " ++ src ++ " to " ++ dest)
   copyFile src dest
 
--- |The path name that represents the current directory.  May be
--- system-specific.  In Unix, it's @\".\"@.
+-- adaptation of removeDirectoryRecursive
+copyDirectoryRecursiveVerbose :: Int -> FilePath -> FilePath -> IO ()
+copyDirectoryRecursiveVerbose verbose srcDir destDir = do
+  when (verbose > 0) $
+    putStrLn ("copy directory '" ++ srcDir ++ "' to '" ++ destDir ++ "'.")
+  let aux src dest =
+         let cp :: FilePath -> IO ()
+             cp f = let srcFile  = joinPaths src  f
+                        destFile = joinPaths dest f
+                    in  do success <- try (copyFileVerbose verbose srcFile destFile)
+                           case success of
+                              Left e  -> do isDir <- doesDirectoryExist srcFile
+                                            -- If f is not a directory, re-throw the error
+                                            unless isDir $ ioError e
+                                            aux srcFile destFile
+                              Right _ -> return ()
+         in  do createDirectoryIfMissing False dest
+                getDirectoryContentsWithoutSpecial src >>= mapM_ cp
+   in aux srcDir destDir
+
+
+
+
+-- | The path name that represents the current directory.
+-- In Unix, it's @\".\"@, but this is system-specific.
+-- (E.g. AmigaOS uses the empty string @\"\"@ for the current directory.)
 currentDir :: FilePath
 currentDir = "."
 
+dirOf :: FilePath -> FilePath
+dirOf f = (\ (x, _, _) -> x) $ (splitFilePath f)
+
 mkLibName :: FilePath -- ^file Prefix
           -> String   -- ^library name.
           -> String
@@ -281,14 +345,21 @@ mkProfLibName :: FilePath -- ^file Prefix
               -> String
 mkProfLibName pref lib = mkLibName pref (lib++"_p")
 
-mkGHCiLibName :: FilePath -- ^file Prefix
-              -> String   -- ^library name.
-              -> String
-mkGHCiLibName pref lib = pref `joinFileName` ("HS" ++ lib ++ ".o")
+-- ------------------------------------------------------------
+-- * Some Paths
+-- ------------------------------------------------------------
+distPref :: FilePath
+distPref = "dist"
 
--- -----------------------------------------------------------------------------
+srcPref :: FilePath
+srcPref = distPref `joinFileName` "src"
+
+haddockPref :: FilePath
+haddockPref = foldl1 joinPaths [distPref, "doc", "html"]
+
+-- ------------------------------------------------------------
 -- * temporary file names
--- -----------------------------------------------------------------------------
+-- ------------------------------------------------------------
 
 -- use a temporary filename that doesn't already exist.
 -- NB. *not* secure (we don't atomically lock the tmp file we get)
@@ -403,14 +474,16 @@ hunitTests
        do mp1 <- moduleToFilePath [""] "Distribution.Simple.Build" suffixes --exists
           mp2 <- moduleToFilePath [""] "Foo.Bar" suffixes    -- doesn't exist
           assertEqual "existing not found failed"
-                   (Just "Distribution\\Simple\\Build.hs") mp1
-          assertEqual "not existing not nothing failed" Nothing mp2,
+                   ["Distribution\\Simple\\Build.hs"] mp1
+          assertEqual "not existing not nothing failed" [] mp2,
 
         "moduleToPossiblePaths 1" ~: "failed" ~:
              ["Foo\\Bar\\Bang.hs","Foo\\Bar\\Bang.lhs"]
                 ~=? (moduleToPossiblePaths "" "Foo.Bar.Bang" suffixes),
         "moduleToPossiblePaths2 " ~: "failed" ~:
               (moduleToPossiblePaths "" "Foo" suffixes) ~=? ["Foo.hs", "Foo.lhs"],
+        TestCase (do files <- filesWithExtensions "." "cabal"
+                     assertEqual "filesWithExtensions" "Cabal.cabal" (head files))
 #else
        do mp1 <- moduleToFilePath [""] "Distribution.Simple.Build" suffixes --exists
           mp2 <- moduleToFilePath [""] "Foo.Bar" suffixes    -- doesn't exist
