@@ -17,8 +17,15 @@ module Control.Concurrent.STM.TVar (
 	TVar,
 	newTVar,
 	readTVar,
-	writeTVar
+	writeTVar,
+	newTVarIO,
+#ifdef __GLASGOW_HASKELL__
+	registerDelay
+#endif
   ) where
 
+#ifdef __GLASGOW_HASKELL__
 import GHC.Conc
-
+#else
+import Control.Sequential.STM
+#endif
